package ch.epfl.collections;


public class HashMap<K, V> implements Map<K, V> {
    private static final double MIN_LOAD_FACTOR = 0.5;
    private static final double IDEAL_LOAD_FACTOR = 0.75;
    private static final double MAX_LOAD_FACTOR = 1;
    private static final int MIN_CAPACITY = 10;
    
    @SuppressWarnings("unchecked")
    private Entry<K, V>[] table = (Entry<K, V>[])new Entry[MIN_CAPACITY]; 
    private int size = 0;
    
    @Override
    public boolean isEmpty() {
        return size == 0;
    }

    @Override
    public int size() {
        return size;
    }

    @Override
    public void put(K key, V value) {
        final int i = getIndex(key);
        final Entry<K, V> n = entryForKey(key, i);
        if (n != null) {
            n.value = value;
        } else {
            table[i] = new Entry<K, V>(table[i], key, value);
            ++size;
            rehashIfNeeded();
        }
    }

    @Override
    public void remove(K key) {
        final int i = getIndex(key);
        Entry<K, V> entry = table[i], prevEntry = null;
        while (entry != null && !entry.key.equals(key)) {
            prevEntry = entry;
            entry = entry.next;
        }
        if (entry != null) {
            if (prevEntry == null)
                table[i] = table[i].next;
            else
                prevEntry.next = entry.next;
            --size;
            rehashIfNeeded();
        }
    }

    @Override
    public V get(K key) {
        Entry<K, V> e = entryForKey(key, getIndex(key));
        return e == null ? null : e.value;
    }

    @Override
    public boolean containsKey(K key) {
        return entryForKey(key, getIndex(key)) != null;
    }

    @Override
    public String toString() {
        StringBuilder b = new StringBuilder("[");
        for (int i = 0; i < table.length; ++i) {
            for (Entry<K, V> e = table[i]; e != null; e = e.next) {
                b.append(e);
                b.append(",");
            }
        }
        return b.replace(b.length() - 1, b.length(), "]").toString();
    }

    private int getIndex(K elem) {
        return Math.abs(elem.hashCode()) % table.length;
    }

    private Entry<K, V> entryForKey(K key, int i) {
        for (Entry<K, V> e = table[i]; e != null; e = e.next) {
            if (e.key.equals(key))
                return e;
        }
        return null;
    }

    @SuppressWarnings("unchecked")
    private void rehashIfNeeded() {
        final double loadFactor = (double)size / (double)table.length;
        if (! ((loadFactor < MIN_LOAD_FACTOR && table.length > MIN_CAPACITY) || (loadFactor > MAX_LOAD_FACTOR)))
            return;
        
        final int newCapacity = Math.max(MIN_CAPACITY, (int) ((double)size / IDEAL_LOAD_FACTOR));
        Entry<K, V>[] newTable = (Entry<K, V>[])new Entry[newCapacity];
        
        for (int i = 0; i < table.length; ++i) {
            for (Entry<K, V> e = table[i]; e != null; e = e.next) {
                final int newIndex = Math.abs(e.key.hashCode()) % newCapacity;
                newTable[newIndex] = new Entry<K, V>(newTable[newIndex], e.key, e.value);
            }
        }
        table = newTable;
    }

    private static class Entry<K, V> {
        public Entry<K, V> next;
        public final K key;
        public V value;

        public Entry(Entry<K, V> next, K key, V value) {
            this.next = next;
            this.key = key;
            this.value = value;
        }

        @Override
        public String toString() {
            return key +"->"+ value;
        }
    }
}
