package ch.epfl.collections;

public class TreeMap<K extends Comparable<K>, V> implements Map<K, V> {
    private Tree<K, V> root = new Leaf<K, V>();

    @Override
    public boolean isEmpty() {
        return root.isEmpty();
    }

    @Override
    public int size() {
        return root.size();
    }

    @Override
    public void put(K key, V value) {
        root = root.put(key, value);
    }

    @Override
    public void remove(K key) {
        root = root.remove(key);
    }

    @Override
    public V get(K key) {
        Node<K, V> n = root.nodeForKey(key);
        return n == null ? null : n.value;
    }

    @Override
    public boolean containsKey(K key) {
        return root.nodeForKey(key) != null;
    }

    @Override
    public String toString() {
        StringBuilder b = new StringBuilder("[");
        root.appendToStringBuilder(b);
        return b.replace(b.length() - 1, b.length(), "]").toString();
    }

    // Tree

    private static interface Tree<K extends Comparable<K>, V> {
        public boolean isEmpty();
        public int size();
        public Tree<K, V> put(K k, V v);
        public Tree<K, V> remove(K k);
        public Node<K, V> nodeForKey(K k);
        public Node<K, V> getMinNode();
        public void appendToStringBuilder(StringBuilder b);
    }

    // Tree node

    private static class Node<K extends Comparable<K>, V> implements Tree<K, V> {
        private final K key;
        private V value;
        private Tree<K, V> s, g;

        public Node(Tree<K, V> s, K key, V value, Tree<K, V> g) {
            this.s = s;
            this.key = key;
            this.value = value;
            this.g = g;
        }

        @Override
        public boolean isEmpty() {
            return false;
        }

        @Override
        public int size() {
            return 1 + s.size() + g.size();
        }

        @Override
        public Tree<K, V> put(K k, V v) {
            int c = k.compareTo(key);
            if (c < 0)
                s = s.put(k, v);
            else if (c > 0)
                g = g.put(k, v);
            else
                value = v;
            return this;
        }

        @Override
        public Tree<K, V> remove(K k) {
            int c = k.compareTo(key);
            if (c < 0) {
                s = s.remove(k);
                return this;
            } else if (c > 0) {
                g = g.remove(k);
                return this;
            } else {
                if (s.isEmpty()) return g;
                else if (g.isEmpty()) return s;
                else {
                    Node<K, V> newRoot = g.getMinNode();
                    return new Node<K, V>(s, newRoot.key, newRoot.value, g.remove(newRoot.key));
                }
            }
        }

        @Override
        public Node<K, V> nodeForKey(K k) {
            int c = k.compareTo(key);
            if (c < 0)
                return s.nodeForKey(k);
            else if (c > 0)
                return g.nodeForKey(k);
            else
                return this;
        }

        @Override
        public Node<K, V> getMinNode() {
            return s.isEmpty() ? this : s.getMinNode();
        }

        @Override
        public void appendToStringBuilder(StringBuilder b) {
            s.appendToStringBuilder(b);
            b.append(key);
            b.append("->");
            b.append(value);
            b.append(",");
            g.appendToStringBuilder(b);
        }
    }

    // Tree leaf

    private static class Leaf<K extends Comparable<K>, V> implements Tree<K, V> {
        @Override
        public boolean isEmpty() {
            return true;
        }

        @Override
        public int size() {
            return 0;
        }

        @Override
        public Tree<K, V> put(K k, V v) {
            return new Node<K, V>(new Leaf<K, V>(), k, v, new Leaf<K, V>());
        }

        @Override
        public Tree<K, V> remove(K k) {
            return this;
        }

        @Override
        public Node<K, V> nodeForKey(K k) {
            return null;
        }

        @Override
        public Node<K, V> getMinNode() {
            throw new UnsupportedOperationException();
        }

        @Override
        public void appendToStringBuilder(StringBuilder b) {
            // nothing to do
        }
    }
}
